/*
    Copyright 2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>
#include "mywav.h"



#define VER     "0.1 modified"
#define FREQUENCY 22050

// parts of code from ffmpeg
typedef struct ADPCMChannelStatus {
    int predictor;
    short int step_index;
    int step;
} ADPCMChannelStatus;

typedef struct ADPCMContext {
    ADPCMChannelStatus status[2];
} ADPCMContext;

static const int index_table[16] = {
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
};

static const int step_table[89] = {
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

int adpcm_decode_init(ADPCMContext *c) {
    memset(c, 0, sizeof(ADPCMContext));
    return 0;
}

short adpcm_ima_expand_nibble(ADPCMChannelStatus *c, char nibble) {
    int     diff,
            step;

    step = step_table[c->step_index];
    c->step_index += index_table[(unsigned)nibble];
    if(c->step_index < 0) c->step_index = 0;
    else if(c->step_index > 88) c->step_index = 88;
    diff = 0;
	if(nibble & 4) diff = step << 2;
	if(nibble & 2) diff += step << 1;
	if(nibble & 1) diff += step;
    diff >>= 2;
    if(nibble & 8) c->predictor -= diff;
    else c->predictor += diff;
    if(c->predictor < -32768) c->predictor = -32768;
    else if(c->predictor > 32767) c->predictor = 32767;
    return((short)c->predictor);
}



int adpcm_decode_frame_file(ADPCMContext *c, int channels, FILE *fd, FILE *fdo, int maxsize) {
    int     st,
            src,
            in_size,
            out_size;
    short   samples[2];

    out_size = 0;
    st = (channels >= 2) ? 1 : 0;

    for(in_size = 0; (src = fgetc(fd)) >= 0; in_size++) {
        if((maxsize >= 0) && (in_size > maxsize)) break;
        samples[0] = adpcm_ima_expand_nibble(&c->status[0], (src >> 4) & 0x0F);
        samples[1] = adpcm_ima_expand_nibble(&c->status[st], src & 0x0F);
        fwrite(&samples, 1, sizeof(samples), fdo);
        out_size += sizeof(samples);
    }

    return(out_size);   // return output size instead of input one
}



int check_overwrite(char *fname) {
    FILE    *fd;
    char    ans[16];

    if(!fname) return(0);
    fd = fopen(fname, "rb");
    if(!fd) return(0);
    fclose(fd);
    printf("- the file already exists, do you want to overwrite it (y/N)? ");
    fgets(ans, sizeof(ans), stdin);
    if(tolower(ans[0]) == 'y') return(0);
    return(-1);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}



int main(int argc, char *argv[]) {
    ADPCMContext    ctx;
    mywav_fmtchunk  fmt;
    struct stat xstat;
    FILE    *fd,
            *fdo;
    int     size;
    char    *in_file,
            *out_file;

    setbuf(stdout, NULL);

    fputs("\n"
        "Omikron ADP -> WAV "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <input.ADP> <output.wav>\n"
            "\n", argv[0]);
        exit(1);
    }

    in_file  = argv[1];
    out_file = argv[2];

    printf("- read file %s\n", in_file);
    fd = fopen(in_file, "rb");
    if(!fd) std_err();
    
    //New Code
    int channels;
    fseek(fd,3,SEEK_SET);
    if(fgetc(fd)) channels = 2;
    else channels = 1;
    printf("- decoding as %d hz - %d channel(s)\n",FREQUENCY,channels);
    //End of new code

    printf("- create file %s\n", out_file);
    if(check_overwrite(out_file) < 0) exit(1);
    fdo = fopen(out_file, "wb");
    if(!fdo) std_err();

    fstat(fileno(fd), &xstat);
    //xstat.st_size -= 16;
    //size = fgetc(fd) | (fgetc(fd) << 8) | (fgetc(fd) << 16);
    //if(size > xstat.st_size) size = xstat.st_size;
    size = xstat.st_size - 16;
    fseek(fd, 16, SEEK_SET);

    fmt.dwSamplesPerSec  = FREQUENCY;
    fmt.wChannels        = channels;
    fmt.wFormatTag       = 1;
    fmt.wBitsPerSample   = 16;
    fmt.wBlockAlign      = (fmt.wBitsPerSample / 8) * fmt.wChannels;
    fmt.dwAvgBytesPerSec = fmt.dwSamplesPerSec * fmt.wBlockAlign;
    mywav_writehead(fdo, &fmt, size * 4, NULL, 0);

    adpcm_decode_init(&ctx);
    adpcm_decode_frame_file(&ctx, fmt.wChannels, fd, fdo, size);

    fclose(fd);
    fclose(fdo);

    printf("- %d bytes written\n", size);
    printf("- done\n");
    return(0);
}


