#include <cstdlib>
#include <iostream>
#include <cstdio>
#include <list>
#include "conio.h"

#define VER "1.0"

using namespace std;

static const int index_table[16] = {
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
};

static const int step_table[89] = {
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

int modulo(int number);
//Luigi's decoding routine with different parameters.
short adpcm_ima_expand_nibble(char nibble, short int step_index, int predictor);
void value_to_file(int value, FILE *stream);
void fatal_error(char message[]);
short word_to_int(unsigned char *pointer);

int main(int argc, char *argv[])
{
             printf("-------------------\n"
                    "Omikron ADP Encoder "VER"\n"
                    "-------------------\n"
                    " by TuxTheWise for the DCRES project\n"
                    "  based on Luigi Auriemma's code.\n"
                    "\n");
    if(argc != 3){
             printf(" Usage:\n"
                    "adp_encoder.exe <input.wav> <output.adp>\n"
                    "\n"
                    "input.wav is assumed being:\n"
                    "PCM signed 16 bit, mono, 22050hz, with first sample starting at 0x2C.\n"
                    "The program won't check if the input is correct.\n");
                    exit(EXIT_SUCCESS);
    }
    printf("Opening input file... ");
    FILE *arq_input = fopen(argv[1],"rb");
    if(!arq_input) fatal_error("Error opening input file.");
    printf("DONE!\n");
    
    printf("Copying input file to memory... ");
    fseek(arq_input,0,SEEK_END);
    int tamanho = ftell(arq_input);
    
    unsigned char *buffer = (unsigned char *)malloc(tamanho-0x2C);
    if(!buffer) fatal_error("Error alocating memory.");
    
    //Skips header.
    fseek(arq_input,0x2C,SEEK_SET);
    int temp = fread(buffer,1,tamanho - 0x2C,arq_input);
    if(temp != (tamanho - 0x2C)) fatal_error("Error reading input file.");
    
    fclose(arq_input);
    printf("DONE!\n");
    
    printf("Encoding... ");
    
    tamanho -= 0x2C;
    
    //Gets PCM samples.
    short *samples = (short *)calloc(tamanho/2,sizeof(short));
    
    for(int i = 0, j = 0; i < tamanho; i += 2, j++){
            samples[j] = word_to_int(&buffer[i]);
    }
    
    tamanho /= 2;
    
    free(buffer);
    
    list<unsigned char> encoded;
    int predicted;
    short step_index;

    
    //ADPCM variables.
    predicted = samples[0];
    step_index = 0;
   
    for(int i = 1; i < tamanho; i++){
            //Tests the 16 nibbles combination, the best (less diference from real sample) is used.
            int diferenca = modulo(samples[i] - adpcm_ima_expand_nibble(0, step_index, predicted));
            int melhor = 0;
            for(int j = 1; j < 16; j++){
                    int testado = adpcm_ima_expand_nibble(j, step_index, predicted);
                    if(modulo(samples[i] - testado) < diferenca){
                            melhor = j;
                            diferenca = modulo(samples[i] - testado);
                    }
            }
            encoded.push_back(melhor);
                    
            //Update variables for the next sample.
            int step = step_table[step_index];
                    
            step_index += index_table[(unsigned)melhor];
            if(step_index < 0) step_index = 0;
            else if(step_index > 88) step_index = 88;
                    
            int delta = melhor & 0x07;
            int magnitude;
            if(melhor &  8) magnitude = -1;
            else magnitude = 1;
    
            int diff = (delta * step) / 16;
            predicted += diff * magnitude;
            if(predicted < -32768) predicted = -32768;
            else if(predicted > 32767) predicted = 32767;
    }
    
    
    
    printf("DONE!\n");
    
    
    printf("Creating output file... ");
    //Ignores last sample if the number of samples is even.
    if(encoded.size() % 2) encoded.pop_back();
    tamanho = encoded.size() / 2;
    
    unsigned char *buffer_output = (unsigned char *)malloc(tamanho);
    if(!buffer_output) fatal_error("Error allocating memory.");
    
    //Mount the bytes (2 samples each) and stores at the final buffer.
    for(int i = 0; i < tamanho; i++){
          int nibble1 = encoded.front();
          encoded.pop_front();
          int nibble2 = encoded.front();
          encoded.pop_front();
          int final_nibble = 0;
          final_nibble |= (nibble1 << 4);
          final_nibble |= nibble2;
          buffer_output[i] = (unsigned char)final_nibble;
    }
    
    printf("DONE!\n");
    
    printf("Writing output file... ");
    
    FILE *arq_output = fopen(argv[2],"wb");
    if(!arq_output) fatal_error("Error creating output file.");
    
    value_to_file(tamanho,arq_output);
    fseek(arq_output,16,SEEK_SET);
    
    temp = fwrite(buffer_output,1,tamanho,arq_output);
    if(temp != tamanho) fatal_error("Error writing in output file.");
    
    fclose(arq_output);
    
    printf("DONE!\n");
    
    printf("\nFinished! %d bytes written.\n",tamanho+16);
    
}

void fatal_error(char message[]){
     printf("\n- %s\n",message);
     exit(EXIT_FAILURE);
}

short word_to_int(unsigned char *pointer){
      
    short output = 0;
    output |= pointer[0];
    output |= (pointer[1]) << 8;
    
    return output;
}

int modulo(int number){
    if(number < 0) number *= -1;
    return number;
}

short adpcm_ima_expand_nibble(char nibble, short int step_index, int predictor) {
    int     diff,
            step;
            
    int delta = nibble & 0x07;
    int magnitude;
    if(nibble &  8) magnitude = -1;
    else magnitude = 1;

    step = step_table[step_index];
    step_index += index_table[(unsigned)nibble];
    if(step_index < 0) step_index = 0;
    else if(step_index > 88) step_index = 88;
    
	diff = (delta * step) / 16;
    predictor += diff * magnitude;
    if(predictor < -32768) predictor = -32768;
    else if(predictor > 32767) predictor = 32767;
    return((short)predictor);
}

void value_to_file(int value, FILE *stream){
     fputc((unsigned char)(value & 0xFF),stream);
     fputc((unsigned char)((value >> 8) & 0xFF),stream);
     fputc((unsigned char)((value >> 16) & 0xFF),stream);
     fputc((unsigned char)((value >> 24) & 0xFF),stream);
}
